/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <GL/glut.h>
#include "types.h"




int contour_levels; /* no of contour levels */



/* contour plotting */
static int 
plot_contour(triag *t, long double potential)
{
  /* variables for contour plotting */
  int H,L,E;
  
  unsigned long int p1,p2,p3;
  long double x1,x2,y1,y2;

#ifdef DEBUG
  printf("contour for triangle %ld\n",t->n);
#endif
  
  
  /* now determine the contours if any */
  /* first determine point potentials */
  E=H=L=0;
  if ( Mpot(t->n1) == potential )
    E++;
  else if ( Mpot(t->n1) < potential )
    L++;
  else
    H++;
  if ( Mpot(t->n2) == potential )
    E++;
  else if ( Mpot(t->n2) < potential )
    L++;
  else
    H++;
  if ( Mpot(t->n3) == potential )
    E++;
  else if ( Mpot(t->n3) < potential )
    L++;
  else
    H++;
  
  /* now according to the values of E,H and L */
  /* fit the general case */
  if ( E == 3 ) { /* 3 contours */
    /* we do not plot this */
#ifdef DEBUG
    glBegin(GL_LINE_LOOP);
    /* get first point */
    glVertex2d( (GLdouble)Mx(t->n1), (GLdouble)My(t->n1));
    /* get second point */
    glVertex2d( (GLdouble)Mx(t->n2), (GLdouble)My(t->n2));  
    /* get third point */
    glVertex2d( (GLdouble)Mx(t->n3), (GLdouble)My(t->n3));
    glEnd();
#endif
#ifdef DEBUG
    printf("3 lines\n");
#endif
    return(3);
  }
  if ( E == 2 ) { /* 1 line */
    /* generalize to p1,p2 */
    if ( Mpot(t->n1) != potential ) {
      p1 = t->n2; p2 = t->n3;
    } 
    else if ( Mpot(t->n2) != potential ) {
      p1 = t->n3; p2 = t->n1;
    } 
    else if ( Mpot(t->n3) != potential ) {
      p1 = t->n1; p2 = t->n2;
    } 
    else
      return(-1); /*error*/
    /* now plot the general line */
     glBegin(GL_LINE_LOOP);
     /* get first point */
     glVertex2d( (GLdouble)Mx(p1), (GLdouble)My(p1));
    /* get second point */
    glVertex2d( (GLdouble)Mx(p2), (GLdouble)My(p2));  
    glEnd();
#ifdef DEBUG
    printf("1 line boundary\n");
#endif
    return(1);
  }
  if ( (E == 1) && ( H == 1 ) && ( L == 1 ) ) {
    /* againe one line generalize */
    if ( Mpot(t->n1) == potential ) {
      p1 = t->n2; p2 = t->n3; p3 = t->n1;
    } 
    else if ( Mpot(t->n2) == potential ) {
      p1 = t->n3; p2 = t->n1; p3 = t->n2;
    } 
    else if ( Mpot(t->n3) == potential ) {
      p1 = t->n1; p2 = t->n2; p3 = t->n3;
    } 
    else 
      return(-1);
    /* now interpolate p1,p2 to find the correct point */
    if ( Mpot(p1) == Mpot(p2) )
      return(-1); /* error */
    x1 = Mx(p2) 
      +( Mx(p1)-Mx(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
    y1 = My(p2)
      +( My(p1)-My(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
    
    /* now draw line between p3,(x1,y1) */
    glBegin(GL_LINE_LOOP);
    /* get first point */
    glVertex2d( (GLdouble)Mx(p3), (GLdouble)My(p3));
    glVertex2d( (GLdouble)x1,(GLdouble) y1 );
    glEnd();
#ifdef DEBUG
    printf("1 line equal\n");
#endif
    return(1);
  }
  
  if (( ( H == 2 ) && ( L ==1 ) ) || ( ( H == 1 ) && ( L == 2 ))) {
    /* two lines, need to generalize */
    
    if ( ( H == 2 ) && ( L ==1 ) ) {
      /* find the lowest point, assign it to p1 */
      if ( Mpot(t->n1) < potential ) {
	p1 = t->n1; p2 = t->n2; p3 = t->n3;
      } else  
      if ( Mpot(t->n2) < potential ) {
	p1 = t->n2; p2 = t->n3; p3 = t->n1;
      } else
      if ( Mpot(t->n3) < potential ) {
	p1 = t->n3; p2 = t->n1; p3 = t->n2;
      } else 
	return(-1); /* error */
#ifdef DEBUG
      printf("1 line case1\n");
#endif
    }
    if ( ( H == 1 ) && ( L == 2 ) ) {
      /* find the highest point, assign it to p1 */
      if ( Mpot(t->n1) > potential ) {
	p1 = t->n1; p2 = t->n2; p3 = t->n3;
      } else
      if ( Mpot(t->n2) > potential ) {
	p1 = t->n2; p2 = t->n3; p3 = t->n1;
      } else
      if ( Mpot(t->n3) > potential ) {
	p1 = t->n3; p2 = t->n1; p3 = t->n2;
      } else
	return(-1);
#ifdef DEBUG
      printf("1 line case2\n");
#endif
    }
    /* now interpolate between (p1,p2) and (p1,p3) */
  
    /* first p1,p2 interpolation */
    if ( Mpot(p1) != Mpot(p2) ) {
      x1 = Mx(p2) 
	+ (Mx(p1)-Mx(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
      y1 = My(p2)
	+ (My(p1)-My(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
    }
    /* now p1,p3 interpolation */
    if ( Mpot(p1) != Mpot(p3) ) {
    x2 = Mx(p3) 
      + (Mx(p1)-Mx(p3))*(potential-Mpot(p3))/(Mpot(p1)-Mpot(p3));
    y2 = My(p3)
      + (My(p1)-My(p3))*(potential-Mpot(p3))/(Mpot(p1)-Mpot(p3));
  
    }
    /* now draw line (x1,y1) (x2,y2) */
    glBegin(GL_LINE_LOOP);
    /* get first point */
    glVertex2d( (GLdouble)x1, (GLdouble) y1 );
    glVertex2d( (GLdouble)x2, (GLdouble) y2 );
    glEnd();
    return(1); 

  }
  
  /* if we reach here, no contour */
  return(0);
}


/* to get a color to plot */
static int 
get_colour(float* redp, float* greenp, float *bluep, int level, int max_levels)
{
  //float cl  = 3.0*(float)level/(float)max_levels;

     
  // if ( cl < 1.5 ) { /* blue - green */
  // *redp = 0.5;
  // *greenp = cl*0.5;
  // *bluep = 1.0 - *greenp;
  //} else  { /* green - red */
  // *redp = 0.25 + (cl-1.5)*0.5;
  // *bluep = 0.0;
  // *greenp = 1.0 - *redp;
  //}
  
  *redp = (float)level/(float)max_levels;
  *greenp = 0.7f;
  *bluep = 1.0f - *redp;
  return(0);

}
  

/* draws legend */
static void
draw_legend(float x1, float y1, float x2, float y2, int levels)
{
  float del_y; /* y step */ 
  float yy1,yy2; /* temp y coords */
  float red,blue,green; /* three colour components */
  char buff[14]; /* buffer for printing */
 
  int i;


  if ( levels <= 0 ) 
    levels = 10;
  
  del_y =( y1 - y2 )/(float)levels;
  
 
  yy1 = del_y + y2;
  yy2 = y2;
  /* draw a horizontal divider (a tick) */
  glColor3f( 0.0f, 0.7f , 0.7f );
  glBegin(GL_LINES);
  glVertex2d( x2,yy2);
  glVertex2d( x2+0.02,yy2);
  glEnd();
  
  for ( i = 0; i < levels ; i++ ) {
  
    get_colour(&red,&green,&blue,i, contour_levels);
    glColor3f(  red, green , blue );
    
    glBegin(GL_QUADS);
    glVertex2d( x1,yy1);
    glVertex2d( x2,yy1);
    glVertex2d( x2,yy2);
    glVertex2d( x1,yy2);
    glEnd();
    
    /* draw a horizontal divider (a tick) */
    glColor3f( 0.0f, 0.7f , 0.7f );
    glBegin(GL_LINES);
    glVertex2d( x2,yy1);
    glVertex2d( x2+0.02,yy1);
    glEnd();
    
    yy1 += del_y;
    yy2+= del_y;
    glFlush();  
  }
  /* print values of max, min potentials */
  sprintf(buff, "%4.2Lf",minpot);
  glRasterPos2d(x2,y2-0.05); 
  printstring(GLUT_BITMAP_HELVETICA_10,buff);
  sprintf(buff, "%4.2Lf",maxpot);
  glRasterPos2d(x2,y1+0.04); 
  printstring(GLUT_BITMAP_HELVETICA_10,buff);
}
  
/* function to plot the contour */
void 
plot_contour_all(void)
{
  triag *t;
  unsigned long int i;
  int j; /* for contour calculation */
  float r,g,b; /* colours */
  long double step =  (maxpot-minpot)/(float)contour_levels; /* contour step */

  
  
#ifdef DEBUG
  printf("plotting from %Lf to %Lf at %Lf intervals\n",maxpot,minpot,step);
#endif

  for ( i = 1; i <= triangle_count; i++ ) {
    t = findTriangle(i);
    if ( t != 0 ) {
      for ( j = 0; j < contour_levels ; j++ ) {
	/* get the colour */
	get_colour(&r,&g,&b,j, contour_levels);
	/* set the colour */
	glColor3f((GLfloat)r, (GLfloat)g, (GLfloat)b);
	plot_contour(t,minpot+j*step);
      }
         
    }
  }

  /* now plot legend */
  draw_legend(-0.95f,0.6f,-0.92f,-0.6f,contour_levels);
}


/* a function to plot the gradient of potential */
void
plot_gradient(void)
{
  triag *t;
  unsigned long int i,n1,n2,n3;
  long double del, a,b, yy12, yy23, yy31, x0,x1,x2,y0,y1,y2;
  
  for ( i = 1; i <= triangle_count; i++ ) {
    t = findTriangle(i);
    if ( t != 0 ) {
#ifdef DEBUG
      printf("plotting gradient for %ld",i);
#endif
      /* calculate gradient */
      /* phi =a*x+b*y+c for each triangle */
      /* need to find a and b only */
      /* using cramer's rule for this */
      yy23 = My(t->n2)-My(t->n3);
      yy31 = My(t->n3)-My(t->n1);
      yy12 = My(t->n1)-My(t->n2);
      del = Mx(t->n1)*yy23 + Mx(t->n2)*yy31 + Mx(t->n3)*yy12;
      if ( del < TOL ) 
	del = -del;
      if ( del < TOL ) 
	continue; /* do not plot */
      /* else */
      del = 0.001/del;
      a = del*(Mpot(t->n1)*yy23 + Mpot(t->n2)*yy31 + Mpot(t->n3)*yy12);
      b = del*(Mx(t->n1)*(Mpot(t->n2)-Mpot(t->n3))
	       +Mx(t->n2)*(Mpot(t->n3)-Mpot(t->n1))
	       +Mx(t->n3)*(Mpot(t->n1)-Mpot(t->n2)));
      
      /* now find the centroid of traingle */
      /* x coord */
      x0 = (Mx(t->n1)+Mx(t->n2)+Mx(t->n3))*0.3333333333333;
      /* y coord */
      y0 = (My(t->n1)+My(t->n2)+My(t->n3))*0.3333333333333;
  
      /* now find the two intersection points of the gradient line */
      /*               n3                                   */
      /*               /\                                   */
      /*            p1/  \                                  */
      /*             /`   \                                 */
      /*            /  `   \                                */
      /*           /    `   \                               */
      /*       n1 -------`----n2                            */
      /*                 p2                                 */
      


      /* at each corner, evaluate line value */
      /* check to see if n1 on line */
      yy12= a*(My(t->n1)-y0) - b*(Mx(t->n1)-x0);
      /* check to see if n2 on line */
      yy23= a*(My(t->n2)-y0) - b*(Mx(t->n2)-x0);
      /* check to see if n2 on line */
      yy31= a*(My(t->n3)-y0) - b*(Mx(t->n3)-x0);

      /* now 2 of the abouve values will have opposite signes from the other one */
      /* so generalize */
      if ( ((yy12 >= TOL) && (yy23 < 0) && (yy31 < 0)) || ((yy12 <= TOL) && (yy23 > 0) && (yy31 > 0))) {
	n1 = t->n1;
	n2 = t->n2;
	n3 = t->n3;
      } else if ( ((yy23 >= TOL) && (yy12 < 0) && (yy31 < 0)) || ((yy23 <= TOL) && (yy12 > 0) && (yy31 > 0))) {
	n1 = t->n2;
	n2 = t->n3;
	n3 = t->n1;
      } else if ( ((yy31 >= TOL) && (yy12 < 0) && (yy23 < 0)) || ((yy31 <= TOL) && (yy12 > 0) && (yy23 > 0)) ) {
	n1 = t->n3;
	n2 = t->n1;
	n3 = t->n2;
      } else 
	continue; /* error */
      
      /* now check intersection between (n1,n2) and (n1,n3) */
      /* first (n1,n2) */
      yy12= a*(My(n1)-y0) - b*(Mx(n1)-x0);
      yy23= a*(My(n2)-y0) - b*(Mx(n2)-x0);
      if ( (yy23 < TOL) && (yy23 > -TOL) ) {
	x1 = Mx(n2);
	y1 = My(n2);
      } else {
	del = -yy12/yy23;
	yy23 = 1/(1+del);
	x1 = (Mx(n1)+del*Mx(n2))*yy23;
	y1 = (My(n1)+del*My(n2))*yy23;
      }
      /* now (n1,n3) */
      yy23= a*(My(n3)-y0) - b*(Mx(n3)-x0);
      if ( (yy23 < TOL) && (yy23 > -TOL) ) {
	x2 = Mx(n3);
	y2 = My(n3);
      } else {
	del = -yy12/yy23;
	yy12 = 1/(1+del);
	x2 = (Mx(n1)+del*Mx(n3))*yy12;
	y2 = (My(n1)+del*My(n3))*yy12;
      }
      
      glColor3f(0.4f, 0.5f+b, 0.5f+a);
      /* now plot gradient line (x1,y1) (x2,y2) */
      glBegin(GL_LINE_LOOP);
      /* get first point */
      glVertex2d( (GLdouble)(x1), (GLdouble)(y1) );
      glVertex2d( (GLdouble)(x2), (GLdouble)(y2) );
      glEnd();
    }
  }
}


 
/* a function to print the contour of a triangle to a postscript file */
static int 
print_contour(FILE *plotf, triag *t, long double potential, float myscale)
{
/* variables for contour plotting */
  int H,L,E;
  
  unsigned long int p1,p2,p3;
  long double x1,x2,y1,y2;
  
  /* note: we use shorthand here as follows:
     /l {lineto} bind def 
     /m {moveto} bind def 
     /n {newpath} bind def 
     /s {stroke} bind def 
     /slw {setlinewidth} bind def 
  */

  
   
#ifdef DEBUG
  printf("contour for triangle %ld\n",t->n);
#endif
  
  
  /* now determine the contours if any */
  /* first determine point potentials */
  E=H=L=0;
  if ( Mpot(t->n1) == potential )
    E++;
  else if ( Mpot(t->n1) < potential )
    L++;
  else
    H++;
  if ( Mpot(t->n2) == potential )
    E++;
  else if ( Mpot(t->n2) < potential )
    L++;
  else
    H++;
  if ( Mpot(t->n3) == potential )
    E++;
  else if ( Mpot(t->n3) < potential )
    L++;
  else
    H++;
  
  /* now according to the values of E,H and L */
  /* fit the general case */
  if ( E == 3 ) { /* 3 contours */
    /* we do not print this */
#ifdef DEBUG
    fprintf(plotf,"n ");
    /* get first point */
    fprintf(plotf,"%Lf %Lf  m ", (xrange+Mx(t->n1)*scale)*myscale, (yrange + My(t->n1)*scale)*myscale);
    /* get second point */
     fprintf(plotf,"%Lf  %Lf  l ", (xrange+Mx(t->n2)*scale)*myscale, (yrange + My(t->n2)*scale)*myscale);
    /* get third point */
     fprintf(plotf,"%Lf %Lf l ", (xrange+Mx(t->n3)*scale)*myscale, (yrange + My(t->n3)*scale)*myscale);
     fprintf(plotf,"cp ");
     fprintf(plotf,"s\n");
#endif
#ifdef DEBUG
    printf("3 lines\n");
#endif
    return(3);
  }
  if ( E == 2 ) { /* 1 line */
    /* generalize to p1,p2 */
    if ( Mpot(t->n1) != potential ) {
      p1 = t->n2; p2 = t->n3;
    } 
    else if ( Mpot(t->n2) != potential ) {
      p1 = t->n3; p2 = t->n1;
    } 
    else if ( Mpot(t->n3) != potential ) {
      p1 = t->n1; p2 = t->n2;
    } 
    else
      return(-1);
    /* now plot the general line */
    fprintf(plotf,"n ");
    /* get first point */
    fprintf(plotf,"%Lf %Lf m ", (xrange+Mx(p1)*scale)*myscale, (yrange + My(p1)*scale)*myscale);
    /* get second point */
    fprintf(plotf,"%Lf %Lf l ", (xrange+Mx(p2)*scale)*myscale, (yrange + My(p2)*scale)*myscale);
    fprintf(plotf,"cp ");
    fprintf(plotf,"s\n");
#ifdef DEBUG
    printf("1 line boundary\n");
#endif
    return(1);
  }
  if ( (E == 1) && ( H == 1) && ( L == 1 ) ) {
    /* againe one line generalize */
    if ( Mpot(t->n1) == potential ) {
      p1 = t->n2; p2 = t->n3; p3 = t->n1;
    } 
    else if ( Mpot(t->n2) == potential ) {
      p1 = t->n3; p2 = t->n1; p3 = t->n2;
    } 
    else if ( Mpot(t->n3) == potential ) {
      p1 = t->n1; p2 = t->n2; p3 = t->n3;
    } 
    else
      return(-1);
    /* now interpolate p1,p2 to find the correct point */
    if ( Mpot(p1) != Mpot(p2) ) { /* check even though impossible */
    x1 = Mx(p2) 
      + (Mx(p1)-Mx(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
    y1 = My(p2)
      + (My(p1)-My(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
    
    /* now draw line between p3,(x1,y1) */
    fprintf(plotf,"n ");
    /* get first point */
    fprintf(plotf,"%Lf %Lf  m ", (xrange+Mx(p3)*scale)*myscale, (yrange + My(p3)*scale)*myscale);
    /* get second point */
    fprintf(plotf,"%Lf  %Lf l ", (xrange+x1*scale)*myscale, (yrange + y1*scale)*myscale);
    fprintf(plotf,"cp ");
    fprintf(plotf,"s\n");
  
#ifdef DEBUG
    printf("1 line equal\n");
#endif
    }
    return(1);
  }
  
  if (( ( H == 2 ) && ( L ==1 ) ) || ( ( H == 1 ) && ( L == 2 ))) {
    /* two lines, need to generalize */
    
    if ( ( H == 2 ) && ( L ==1 ) ) {
      /* find the lowest point, assign it to p1 */
      if ( Mpot(t->n1) < potential ) {
	p1 = t->n1; p2 = t->n2; p3 = t->n3;
      } else 
      if ( Mpot(t->n2) < potential ) {
	p1 = t->n2; p2 = t->n3; p3 = t->n1;
      } else
      if ( Mpot(t->n3) < potential ) {
	p1 = t->n3; p2 = t->n1; p3 = t->n2;
      } else
	return(-1); /* error */
#ifdef DEBUG
      printf("1 line case1\n");
#endif
    }
    if ( ( H == 1 ) && ( L == 2 ) ) {
      /* find the highest point, assign it to p1 */
      if ( Mpot(t->n1) > potential ) {
	p1 = t->n1; p2 = t->n2; p3 = t->n3;
      } else
      if ( Mpot(t->n2) > potential ) {
	p1 = t->n2; p2 = t->n3; p3 = t->n1;
      } else
      if ( Mpot(t->n3) > potential ) {
	p1 = t->n3; p2 = t->n1; p3 = t->n2;
      } else
	return(-1);
#ifdef DEBUG
      printf("1 line case2\n");
#endif
    }
    /* now interpolate between (p1,p2) and (p1,p3) */
  
    /* first p1,p2 interpolation */
    if ( Mpot(p1) != Mpot(p2) ) {
      x1 = Mx(p2) 
	+ (Mx(p1)-Mx(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
      y1 = My(p2)
	+ (My(p1)-My(p2))*(potential-Mpot(p2))/(Mpot(p1)-Mpot(p2));
    } else
      return(-1);
    /* now p1,p3 interpolation */
    if ( Mpot(p1) != Mpot(p3) ) {
    x2 = Mx(p3) 
      + (Mx(p1)-Mx(p3))*(potential-Mpot(p3))/(Mpot(p1)-Mpot(p3));
    y2 = My(p3)
      + (My(p1)-My(p3))*(potential-Mpot(p3))/(Mpot(p1)-Mpot(p3));
    } else
      return(-1);
    /* now draw line (x1,y1) (x2,y2) */
    fprintf(plotf,"n ");
    /* get first point */
    fprintf(plotf,"%Lf %Lf m ", (xrange+x1*scale)*myscale, (yrange + y1*scale)*myscale);
    /* get second point */
    fprintf(plotf,"%Lf %Lf  l ", (xrange+x2*scale)*myscale, (yrange + y2*scale)*myscale);
    fprintf(plotf,"cp ");
    fprintf(plotf,"s\n");
    return(1); 

  }
  
  /* if we reach here, no contour */
  return(0);
}

/* a function to generate postscript output of the contour */
void 
print_contour_all(void)
{
  FILE *plotf; /* plot file */
  triag *t;
  unsigned long int i;
  int j; /* for contour calculation */
  long double myscale;
  long double step = (maxpot-minpot)/(float)contour_levels; /* contour step */
  

  plotf =fopen( "cont.eps", "w" );
  if ( plotf == NULL ) {
    fprintf(stderr, "could not open file %s\n", "cont.eps");
    exit(1);
  }

  /* determine the right bound of bounding box */
  if ( xrange > yrange )
    myscale = 400.0/xrange;
  else
    myscale = 400.0/yrange;

  /* print the header */
  fprintf(plotf,"%%!PS-Adobe-2.0 EPSF-3.0\n");
  fprintf(plotf,"%%%%Title: cont.eps\n");
  fprintf(plotf,"%%%%Generated By: "PACKAGE" "VERSION"  contour levels: %d, min= %Lf, max= %Lf \n", contour_levels, minpot, maxpot);
  /* now bounding box -add border of 20 */
  fprintf(plotf,"%%%%BoundingBox: 0 0 %4.2Lf %4.2Lf\n",2*xrange*myscale+20, 2*yrange*myscale+20 );
  fprintf(plotf,"%%%%Magnification: 1.0000\n");

  
  /* set up some font to print text */
  //fprintf(plotf,"/Times-Roman findfont\n");
  //fprintf(plotf,"10 scalefont\n");
  //fprintf(plotf,"setfont\n");
  //fprintf(plotf,"newpath\n");
  // fprintf(plotf,"190 700 moveto\n");
  

  /* start plotting from 10 10 */
  fprintf(plotf,"10 10 translate\n");


  /* now print some shorter definitions */
  fprintf(plotf,"/l {lineto} bind def\n"); /* lineto */
  fprintf(plotf,"/m {moveto} bind def\n"); /* moveto */
  fprintf(plotf,"/n {newpath} bind def\n"); /* newpath */
  fprintf(plotf,"/s {stroke} bind def\n"); /* stroke */
  fprintf(plotf,"/slw {setlinewidth} bind def\n"); /* setlinewidth */
  fprintf(plotf,"/cp {closepath} bind def\n"); /* closepath */
  
  /* now print contours */
  fprintf(plotf,"1 slw\n");
  for ( i = 1; i <= triangle_count; i++ ) {
    t = findTriangle(i);
    if ( t ) {
      for ( j = 0; j < contour_levels; j++ ) 
      	print_contour(plotf,t,minpot+j*step, myscale);
    }
  }
   
  /* then plot the boundaries */
  /* set the line thikness */
  fprintf(plotf,"2 slw\n");
  for ( i = 0; i < bedges; i++ ) {
    fprintf(plotf,"n ");
    fprintf(plotf,"%Lf %Lf  m ",(Mx(edgearray[i].p1)*scale+xrange)*myscale, (My(edgearray[i].p1)*scale+yrange)*myscale);
    fprintf(plotf,"%Lf %Lf  l ",(Mx(edgearray[i].p2)*scale+xrange)*myscale, (My(edgearray[i].p2)*scale+yrange)*myscale);
    fprintf(plotf,"s\n");  
  }
  fprintf(plotf,"%%%%EOF\n");
  
  fclose(plotf);
}
