/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <GL/glut.h>
#include <stdlib.h> /* for strtod() */
#include <stdio.h>
#include <math.h>
#include "types.h"

/* global varibles */
int windowid;
int subwindowid=-1;

#ifdef  D3VIEW
float viewx = 1.0;
float viewy = 1.0;
float viewz = 0.0;
#endif /* !D3VIEW */

/* input data file */
char *infile;

/* global variable needed for drawing mode */
/* status message */
char status[100] = "mouse [right button: menu left button: zoom]";


option opt = BOUND; /* initially draw the boundary and mesh*/
int zoomstate=0; /* flag for zoomstate */
/* zoomstate == 0 NO zoom, zoomstate == 1 zoom ON */ 
/* current cursor coordinatres */
long double zx1,zy1,zx2,zy2; /* for zooming */

static void 
Reshape(int width, int height)
{
  glViewport(0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(-1, 1, -1, 1);
  glMatrixMode(GL_MODELVIEW);
}

/* font rendering-- use stroked fonts */
void 
printstring( void * font, char * str )
{
  int i, l=strlen(str);
  for ( i = 0; i < l; i++ )
    glutBitmapCharacter(font, *str++);
}


#ifdef EXTRA_SOURCE
static void    
Reshape_3d( int w, int h)
{
        glViewport(0,0, (GLsizei) w, (GLsizei) h);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
//      glFrustum(-1.0,1.0,-1.0,1.0,1.5,20.0);
        gluPerspective(60.0,1.0,1.5,20.0);
        glMatrixMode(GL_MODELVIEW);
}
#endif /* !EXTRA_SOURCE */


/* the main drawing function */
static void 
Draw( void )
{
  unsigned long int i;
  
  char buff[50]; /* need for text */
  
  glClear(GL_COLOR_BUFFER_BIT);
  glLineWidth(1.0);
  
  /* for double buffering */
  glPushMatrix();
 
 /* print some text too */

  /* printing this gives color problems */
  
  glColor3f(0.9f,0.7f,0.5f);
  if ( zoomstate == 1 )
    sprintf(status,"zoom window");
  else
    sprintf(status,"mouse [right button: menu left button: zoom]");
  sprintf(buff, status);
  glRasterPos2d(-1.0,0.9); 
  printstring(GLUT_BITMAP_9_BY_15,buff);
  
  switch ( opt ) {
 
  case MESH :
    glColor3f(0.0f, 0.3f, 1.0f); 
    visitAndDrawTree();
    break;


  case GRAD :
    // glColor3f(0.5f,0.5f,0.9f);
    plot_gradient();
    break;


  case ALL :
    glColor3f(0.0f, 1.0f, 1.0f); 
    glBegin(GL_POINTS);
    for ( i = 0; i < npoints; i++ )
      glVertex2d( (GLdouble)Mx(i), (GLdouble)My(i)); 
    glEnd();
    glColor3f(0.0f, 0.3f, 1.0f);
    visitAndDrawTree();
    plot_gradient();
    plot_contour_all();
    break;
 
     case BOUND :
     /* plot contour */
    plot_contour_all(); 
#ifdef NUMBER
    for ( i = 0; i < M.count; i++ ) {
      glColor3f(0.3f, 1.0f-0.05*Mpot(i) ,0.4f+0.05*Mpot(i) );
      glBegin(GL_POINTS);
      glVertex2d( (GLdouble)Mx(i), (GLdouble)My(i) );
      glEnd();
      /* plot the point number */
      if( Mval(i) == FX )
	sprintf(buff,"%ldF",i);
      else
	sprintf(buff,"%ldV",i);
      if ( (M.Narray[i])->pos == IN )
	glColor3f(0.9f, 0.5f, 0.6f);
      else 
	glColor3f(0.0f, 1.0f, 1.0f);
      glRasterPos2d(Mx(i), My(i));
      printstring(GLUT_BITMAP_HELVETICA_10,buff);
    }
#endif
    break;
    
  default:
    break;
  }
  
  /* then plot boundaries */
  glBegin(GL_LINES);
  for ( i = 0; i < bedges; i++ ) {
    if ( edgearray[i].type == NU )
      glColor3f(0.1f, 1.0f, 0.2f);
    else 
      glColor3f(0.9f, 0.4f, 0.9f);
    glVertex2d( (GLdouble)Mx(edgearray[i].p1), (GLdouble)My(edgearray[i].p1));
    glVertex2d( (GLdouble)Mx(edgearray[i].p2), (GLdouble)My(edgearray[i].p2));
  }
  glEnd();
  glFlush();
  glutPostRedisplay();
  glPopMatrix();
  glutSwapBuffers();
 
  
}

/*printing help text */
static void 
Draw_Help(void)
{
  char buff[50];
  
  glClear(GL_COLOR_BUFFER_BIT);
  glutSetCursor(GLUT_CURSOR_DESTROY);
  glColor3f(0.0f,1.0f,0.4f); 
  sprintf(buff, PACKAGE" "VERSION);
  glFlush();  
  glRasterPos2d(-0.5,0.5); printstring(GLUT_BITMAP_TIMES_ROMAN_24,buff);
  glColor3f(1.0f,0.2f,0.4f); 
  sprintf(buff, "Sarod Yatawatta <sarod@cs.pdn.ac.lk>");
  glRasterPos2d(-0.95,0.0); printstring(GLUT_BITMAP_8_BY_13,buff);
  glFlush(); 
  glColor3f(1.0f,0.2f,0.8f); 
  sprintf(buff, "Copyright (c) 2001");
  glRasterPos2d(-0.55,-0.4); printstring(GLUT_BITMAP_9_BY_15,buff);
  glFlush();
}

/* for help window */
static void 
Help_MenuHandler(int option)
{
  
  if ( option == 8 ) {
    glutDestroyWindow( subwindowid );
  }
  subwindowid = -1;
}

/* a function to print help screen */
static void 
print_help(void)
{
 
  glutInitWindowSize(300, 200);
  glutInitWindowPosition(200, 200);
  
  glutInitDisplayMode(GLUT_RGB);
  if ( subwindowid == -1 )
    subwindowid=glutCreateWindow("About..");
  glutCreateMenu(Help_MenuHandler);
  glutAddMenuEntry("close",8);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
  /* observe mouse functions  using the menu */
  
  glutReshapeFunc(Reshape);
  glutDisplayFunc( Draw_Help );
  glClearColor(0, 0, 0, 0);
  glutMainLoop(); 
}

/* function to generate an encapsulated postscript output of the mesh */
static void 
print_mesh(void)
{

  FILE *plotf; /* plot file */
  triag *t;
  unsigned long int i;
  long double myscale;

  plotf =fopen( "mesh.eps", "w" );
  if ( plotf == NULL ) {
    fprintf(stderr, "could not open file %s\n", "mesh.ps");
    exit(1);
  }

  /* determine the right bound of bounding box */
  if ( xrange > yrange )
    myscale = 400.0/xrange;
  else
    myscale = 400.0/yrange;

  /* print the header */
  fprintf(plotf,"%%!PS-Adobe-2.0 EPSF-3.0\n");
  fprintf(plotf,"%%%%Title: mesh.eps\n");
  fprintf(plotf,"%%%%Generated By: "PACKAGE" "VERSION"  triangles= %ld points: initial= %ld final= %ld\n",triangle_count, npoints, M.count );
  /* now bounding box -add border of 20 */
  fprintf(plotf,"%%%%BoundingBox: 0 0 %4.2Lf %4.2Lf\n",2*xrange*myscale+20, 2*yrange*myscale+20 );
  fprintf(plotf,"%%%%Magnification: 1.0000\n");
  
  /* start plotting from 10 10 */
  fprintf(plotf,"10 10 translate\n");

  
  /* now print some shorter definitions */
  fprintf(plotf,"/l {lineto} bind def\n"); /* lineto */
  fprintf(plotf,"/m {moveto} bind def\n"); /* moveto */
  fprintf(plotf,"/n {newpath} bind def\n"); /* newpath */
  fprintf(plotf,"/s {stroke} bind def\n"); /* stroke */
  fprintf(plotf,"/slw {setlinewidth} bind def\n"); /* setlinewidth */
  fprintf(plotf,"/cp {closepath} bind def\n"); /* closepath */
  
  

  /* first plot the triangulation */
  fprintf(plotf,"1 slw\n");
  for ( i = 1; i <= triangle_count; i++ ) {
    t =(triag*) findTriangle(i);
    if ( t ) {
      fprintf(plotf,"n ");
      fprintf(plotf,"%Lf %Lf m ", (xrange+Mx(t->n1)*scale)*myscale,(yrange+My(t->n1)*scale)*myscale);
      fprintf(plotf,"%Lf %Lf l ", (xrange+Mx(t->n2)*scale)*myscale,(yrange+My(t->n2)*scale)*myscale);
      fprintf(plotf,"%Lf %Lf l ", (xrange+Mx(t->n3)*scale)*myscale,(yrange+My(t->n3)*scale)*myscale);
      fprintf(plotf,"cp ");
      fprintf(plotf,"s\n");
    }
  }
  /* then plot the boundaries */
  /* set the line thikness */
  fprintf(plotf,"2 slw\n");
  for ( i = 0; i < bedges; i++ ) {
    fprintf(plotf,"n ");
    fprintf(plotf,"%Lf %Lf  m ",(Mx(edgearray[i].p1)*scale+xrange)*myscale, (My(edgearray[i].p1)*scale+yrange)*myscale);
    fprintf(plotf,"%Lf %Lf  l ",(Mx(edgearray[i].p2)*scale+xrange)*myscale, (My(edgearray[i].p2)*scale+yrange)*myscale);
    fprintf(plotf,"s\n");  
  }
  fprintf(plotf,"%%%%EOF\n");
 
  fclose(plotf);
  
}



/* create a menu */
static void 
MenuHandler(int option)
{
  
  if ( option == 6 ) {
    do {
      delaunay_refine();
      printf("solving..\n");
      re_number_and_solve();
      printf("updating..\n");
      update_priorities();
      printf("dividing bad..\n");
    } while(  divide_bad_triangles() );
    delaunay_refine();
  }
  if ( option == 7 ) 
    if ( subwindowid == -1 )
      print_help();
  if ( option == 8 ) {
    // printf("time to quit\n");
    /* now print the triangulation */
#ifdef DEBUG
    visitTree(root);
#endif
    glutDestroyWindow( windowid );
    printf("triangles %ld points %ld total %ld\n", triangle_count, npoints, M.count);
    print_mesh();
    print_contour_all();
    exit(0);
  }
  if ( option == 2 ) {
    //printf("resize\n");
    glutReshapeWindow(400, 400);
    glutPositionWindow(100, 100);
    glutPostRedisplay();
  } 
  if ( option == 1 ) {
    //printf("full size\n");
    glutFullScreen(); /* occupy full screen */ 
    glutPostRedisplay();
  }
}

/* submenu - drawing */
static void 
SubMenuHandler( int option )
{
  switch( option ) {
  case 1:
    opt = BOUND;
    break;
  case 2:
    opt = MESH;
    break;
  case 3:
    opt = GRAD;
    break;
  case 4:
    opt = ALL;
    break;
  default:
    opt = BOUND;
  }
}

/* mouse callback */
static void 
mouse_handler(int button, int state, int h, int v )
{
  
  unsigned int mx,my;
  if ( button == GLUT_LEFT_BUTTON && state == GLUT_DOWN ) {
    /* first change cursor */
    if ( zoomstate == 0 ) { /* initially OFF */
      glutSetCursor(GLUT_CURSOR_FULL_CROSSHAIR); /* change cursor */
      zoomstate = 1; /* turn ON */
    } else if ( zoomstate == 2 ) { /* we are in a zoom !*/
      /* get back to original mode */
      /* reverse the translation and scaling */
      /* first the translation, and then the scale */
      /* but again we do it the other way! */
      glTranslatef(0.5*(zx1+zx2),0.5*(zy1+zy2),0);
      glScalef(-0.5*(zx1-zx2),0.5*(zy1-zy2),1.0);
      zoomstate = 0;
    } else { /* zooming about to start */ 
      glutSetCursor(GLUT_CURSOR_INHERIT); /* change back to normal */
      zoomstate = 2;    /* OFF */
    }
    /* add new point */
    mx = glutGet(GLUT_WINDOW_WIDTH);
    my = glutGet(GLUT_WINDOW_HEIGHT);
    
    //  printf("adding point %d , %d\n", h, v);

    if ( zoomstate == 1 ) {
    zx1 = (2*( long double )h/( long double ) mx-1.0);
    zy1 = -2*( long double )v/( long double ) my+1.0;
    } else if ( zoomstate == 2 ) { /* need to zoom */
    /* if zoomstate== 0 now, it was initially 1 */
    /* so, need to create a zoom window */
      zx2 = (2*( long double )h/( long double ) mx-1.0);
      zy2 = -2*( long double )v/( long double ) my+1.0;
      if ( zx1 != zx2 ) {
      /* here we do the scaling+ translation in the opposite order */
      /* then the scaling */
      glScalef(-2.0/(zx1-zx2),2.0/(zy1-zy2),1.0);
      /* first the translation */
      glTranslatef(-0.5*(zx1+zx2),-0.5*(zy1+zy2),0);
      }
    }
#ifdef DEBUG
    printf("x1= %Lf y1= %Lf x2= %Lf y2= %Lf  \n",zx1,zy1,zx2,zy2);
#endif
  }
  glutPostRedisplay();
}

#ifdef D3VIEW
static void 
mouse_handler_3d( int h, int v )
{
  unsigned int mx,my;
  static int oldh, oldv;
  
  mx = glutGet(GLUT_WINDOW_WIDTH);
  my = glutGet(GLUT_WINDOW_HEIGHT);
  
  viewx += (float)(h-oldh)*2/(float)mx;
  viewy += (float)(v-oldv)*2/(float)my;
  viewz += (float)(v-oldv+h-oldh)*2/(float)my;
  
  oldh = h;
  oldv = v;
  glutPostRedisplay();
}

#endif


/* function to print help */
static void 
print_info(void)
{
  printf("usage: "PACKAGE" [options] -i inputfile \n");
  printf("options are\n");
  printf("-h, -?: help, print info\n");
  printf("-v    : be verbose\n");                                         
  printf("-a (val) : set min area of each triangle to val\n");                             
  printf("-b (val) : set max badness of each triangle to val\n");            
  printf("-s (val) : set splitting priority to val\n");                 
  printf("-c (val) : set contour levels to val\n");                     
  printf("-nx      : do not go into display mode, dump output to file\n"); 
  printf("-do      : output data file of mesh as \"data.in\"\n");     
  printf("\n\n");
  printf(PACKAGE"  "VERSION", Copyright (c) 2001 Sarod Yatawatta\n");
  printf(PACKAGE" comes with ABSOLUTELY NO WARRANTY\n"); 
  printf("This is free software, and you are welcome to redistribute\n");
  printf("it under certain conditions given by GNU GENERAL PUBLIC LICENSE\n");
  
}

int main(int argc, char *argv[])
{
  int submenu;
  int idx, verbose, xmode, dataout; /* for input parsing */
  
  /* setup globals */
  maximum_area = 0.01; /* largest area of a triangle */
  badness_limit = 5; /* miimum badness for splitting */
  splitter = 0.05; /* to set triangle priority */
  contour_levels = 30;

 
  /* default values */
  infile = 0; /* no */
  verbose = 0; /* no */
  xmode = 1; /* yes */
  dataout =0; /* no */

  /* get standard input */
  /* format: command max_area splitting_priority contour_levels */
  /* format: command [options]*                                 */
  /* options are:                                               */
  /* -h, -?: help, print info                                   */
  /* -v    : be verbose                                         */
  /* -a (val) : set max area to val                             */
  /* -s (val) : set splitting priority to value                 */
  /* -c (val) : set contour levels to value                     */
  /* -nx      : do not go into display mode, dump output to file */
  /* -i (val) : set input file to "value"                        */
  /* -do      : output data file of mesh as "data.in"            */
  
  idx = 1;
  while ( idx < argc-1 ) {
    if ( argv[idx][0] != '-' ) {
      print_info();
      return(0);
    }
    if ( (strcmp(argv[idx],"-?") == 0) || (strcmp(argv[idx], "-h") == 0) ) {
      print_info();
      return(0);
    }
    if ( strcmp(argv[idx],"-v")==0 ) {
      verbose = 1;
      idx++;
     
    }
    else if ( strcmp(argv[idx],"-nx")==0 ) {
      xmode = 0;
      idx++;
     
    }
    else if ( strcmp(argv[idx],"-do")==0 ) {
      dataout = 1;
      idx++;
     
    }
    else if ( strcmp(argv[idx],"-a") == 0 ) {
      if ( idx+1 <= argc-1 ) {
	maximum_area =strtod(argv[idx+1],0);
	if ( maximum_area == 0.0f ) {
	  print_info();
	  maximum_area = 0.01;
	  idx += 1;
	} else 
	  idx += 2;
      }
     
    }
    else if ( strcmp(argv[idx],"-b") == 0 ) {
      if ( idx+1 <= argc-1 ) {
	badness_limit =strtod(argv[idx+1],0);
	 if (  badness_limit == 0.0f ) {
	  print_info();
	  badness_limit = 1.1;
	  idx += 1;
	} else 
	  idx += 2;
      }
     
    }
	      
    else if ( strcmp(argv[idx],"-c") == 0 ) {
      if ( idx+1 <= argc-1 ) {
	contour_levels = strtol(argv[idx+1],0,0);
	if (  contour_levels  == 0.0f ) {
	  print_info();
	  contour_levels  = 30;
	  idx += 1;
	} else 
	  idx += 2;
      }
     
    }
    else if ( strcmp(argv[idx],"-s") == 0 ) {
      if ( idx+1 <= argc-1 ) {
	splitter =strtod(argv[idx+1],0);
	if ( splitter == 0.0f ) {
	  print_info();
	  splitter = 0.05;
	  idx += 1;
	} else 
	  idx += 2;
      }
     
    }
    else if ( strcmp(argv[idx],"-i") == 0 ) {
      if ( idx+1 <= argc-1 ) {
	infile = (char*)malloc((strlen(argv[idx+1])+1)*sizeof(char));
	if ( infile )
	  strcpy(infile,argv[idx+1]);
	else {
	  perror("not enough memory");
	  exit(1);
	}
	idx += 2;
      } else {
	printf("no filename given\n");
	print_info();
	exit(1);
      } 
      
    }
    else { /* unknown option */
      print_info();
      exit(1);
    }
  }
  
  if ( !infile ) {
    printf("no input filename!\n");
    print_info();
    exit(1);
  }


  
  /* print some info */
  printf(PACKAGE" "VERSION", Copyright (c) 2001 Sarod Yatawatta\n");
  printf(PACKAGE" comes with ABSOLUTELY NO WARRANTY\n"); 
  printf("This is free software, and you are welcome to redistribute\n");
  printf("it under certain conditions given by GNU GENERAL PUBLIC LICENSE\n");
  if ( verbose )
    printf("params: input=%s area=%Lf split=%Lf levels=%d  verbose=%d xmode=%d\n", infile ,  maximum_area, splitter,  contour_levels, verbose, xmode);
  
  delaunay();
  if ( verbose )
    printf("removing crossings..\n");
  while(remove_crossings());
  remove_crossings();
  if ( verbose )
    printf("deleting outer triangles..\n");
  cleanup_exterior();
  if ( verbose )
    printf("dividing bad..\n");
  divide_bad_triangles();
  divide_corners();
  while(  divide_bad_triangles() )
    delaunay_refine();
  if ( verbose )
    printf("dividing corners..\n");
  do {
    delaunay_refine();
    if ( verbose )
      printf("solving..\n");
    re_number_and_solve();
    if ( verbose )
      printf("updating..\n");
    update_priorities();
    if ( verbose )
      printf("dividing bad..\n");
  } while(  divide_bad_triangles() );
  delaunay_refine();
  /* mesh related things */
#ifdef DEBUG
  visit(M);
#endif
  
  /* from now on, only the mesh can be used */

  if (dataout ) /* output data file of mesh */
    write_mesh_file("data.in");
  if ( !xmode ) { /* no display mode */
    printf("triangles %ld points %ld total %ld\n", triangle_count, npoints, M.count);
    print_mesh();
    print_contour_all();
    exit(0);
  }

  glutInit(&argc, argv);
  glutInitWindowSize(400, 400);
  glutInitWindowPosition(100, 100);
  
#ifndef D3VIEW
  glutInitDisplayMode(GLUT_RGB|GLUT_DOUBLE);
  windowid=glutCreateWindow(PACKAGE" "VERSION" - 2001");
   /* create a menu and a submenu too!*/
  submenu= glutCreateMenu(SubMenuHandler);
  glutAddMenuEntry("contours",1);
  glutAddMenuEntry("triangles",2);
  glutAddMenuEntry("gradient",3);
  glutAddMenuEntry("all",4);
   /* main menu */
  glutCreateMenu(MenuHandler);
  glutAddMenuEntry("Full Screen", 1);
  glutAddMenuEntry("Resize",2);
  glutAddSubMenu("Draw Mode", submenu);
  glutAddMenuEntry("Solve",6);
  glutAddMenuEntry("About..",7);
  glutAddMenuEntry("Exit",8);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
  /* observe mouse functions  using the menu */
  
  /* mouse handler */
  glutMouseFunc( mouse_handler ) ;
  /* passive motion handler */
  // glutPassiveMotionFunc( mouse_passive );

  glutReshapeFunc(Reshape);
  glutDisplayFunc( Draw );
  glClearColor(0, 0, 0, 0);
  
  
#endif /* not 3DVIEW */
#ifdef D3VIEW
  glutInitDisplayMode(GLUT_RGB|GLUT_SINGLE);
  windowid=glutCreateWindow(PACKAGE" "VERSION" - 2001");
  glutReshapeFunc(Reshape_3d);
  glutDisplayFunc(Draw_3d);
  glutMotionFunc( mouse_handler_3d ) ;
  glClearColor(0, 0, 0, 0);
#endif /* 3DVIEW */
  
  glutMainLoop(); 
  return(0);
}
