/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include "types.h"



/* we need a different structure here */
typedef struct line_ {
  point p1;
  point p2;
}line;



/* a function to check three points are counter clockwise */
static int 
ccw( point p0,point p1,point p2 )
{
  long double dx1, dx2, dy1, dy2, dxy12, dxy21;
  dx1=p1.x-p0.x; dy1=p1.y-p0.y;
  dx2=p2.x-p0.x; dy2=p2.y-p0.y;
  dxy12=dx1*dy2; dxy21=dx2*dy1;

  if ( dxy12 > dxy21 )/* grad p2p0 > grad p1p0 */
  return( 1 );
  if ( dxy12 < dxy21 )  /* grad p2p0 < grad p1p0 */
  return( -1 );
  /* will not reach here if any of above is true */
  /* so here dxy12 == dxy21 i.e. colinear points */
  if (( dx1*dx2 < TOL )||(dy1*dy2 < TOL))/* p0 inbetween p1 and p2 */
  return( -1 );
  if (( dx1*dx1 + dy1*dy1 ) >= ( dx2*dx2 + dy2*dy2 ))
    return(0);
  /* else */
  return(1);
}


/* a function to check the intersection of two lines */
/* __ LOCAL __ */
static int 
intersectl(line l1,line l2)
{

  if ( ( (ccw(l1.p1, l1.p2, l2.p1)*ccw(l1.p1,l1.p2,l2.p2))<= 0 ) &&
	 ( (ccw(l2.p1, l2.p2, l1.p1)*ccw(l2.p1,l2.p2,l1.p2))<= 0 ) )
#ifdef DEBUG
    {
      printf(" yes\n");
#endif
      return(1);
#ifdef DEBUG
    }
  printf(" no\n");
#endif
  return(0);
}

/* a function to check whether a point p is on the line segment */
/* __ LOCAL __ */
#ifdef EXTRA_SOURCE
static int 
on_linel(line l, point p)
{
 long double dx1, dx2, dy1, dy2, dxy12, dxy21;
  dx1=l.p1.x-p.x; dy1=l.p1.y-p.y;
  dx2=l.p2.x-p.x; dy2=l.p2.y-p.y;
  dxy12=dx1*dy2; dxy21=dx2*dy1;

  /* so here dxy12 == dxy21 i.e. colinear points */
  if ((dxy12 == dxy21 )&&(( dx1*dx2 < TOL )||(dy1*dy2 < TOL)))
    /* p inbetween p1 and p2 */
    return(1);
  if((( dx1 <= TOL )&&( dy1 <= TOL)) || (( dx2 <= TOL )&& ( dy2 <= TOL)))
     /* p coincide with one point */
    return(1);
  return(0);
}
#endif /* ! EXTRA_SOURCE */

/* finally, the function to check the inclusion of a point in the polygon */
/* the array has the edge numbers, all edgecount in number */
static position 
inside( point t, unsigned long int edgecount, unsigned long * array )
{
  int count, i, j;
  line lt, lp;
  
  unsigned long int *ta; /* temporary array */
  
  /* assume that array has edgecount number of variables */
  
  /* first do an initial loop to check whether the point is on any boundary */
  /*  NO NEED TO CHECK THIS */
  /* CORRECTION : I think weneed to check this */
#ifdef DEBUG1
  for ( i = 0; i < edgecount; i++) {
    lp.p1.x= Mx(edgearray[array[i]].p1);
    lp.p1.y = My(edgearray[array[i]].p1);

    lp.p2.x= Mx(edgearray[array[i]].p2);
    lp.p2.y = My(edgearray[array[i]].p2);

    if (on_linel(lp,t)) { /*printf("pre detect\n");*/
    return(IN);}
  }
#endif

   /* now copy the node numbers of array to temporary array */
  /* NOTE: in ccw order */
  /* first malloc */
  ta = (unsigned long int*)calloc(edgecount+1, sizeof(unsigned long int));
  
  /* now copy */
  if ( edgearray[array[0]].p1 != edgearray[array[1]].p1
       && edgearray[array[0]].p1 != edgearray[array[1]].p2 )
    ta[0] = edgearray[array[0]].p1;
  else
    ta[0]  = edgearray[array[0]].p2;
  for ( i = 1; i < edgecount; i++ ) {
    if ( edgearray[array[i]].p1 ==  edgearray[array[i-1]].p1 
	 || edgearray[array[i]].p1 ==  edgearray[array[i-1]].p2 )
      ta[i] = edgearray[array[i]].p1;
    else
      ta[i] = edgearray[array[i]].p2;
  }

  ta[i] = ta[0];
#ifdef DEBUG
  printf("\n[");
  for ( i = 0 ; i < edgecount; i++ )
    printf("(%ld, %ld), ",edgearray[array[i]].p1,edgearray[array[i]].p2);
  printf("]\n");
  printf("[");
  for ( i = 0 ; i <= edgecount; i++ )
    printf(" %ld, ",ta[i]);
  printf("]\n");
#endif
  lt.p1=t;
  count = 0 ;
  j = 0;

  /* the test line: ideally this should be a random value */
  lt.p2.x= -100000000+edgecount*1000; /* a very high value */
  lt.p2.y= 100000000-edgecount*1000; 

  for ( i = 1; i < edgecount; i++ ) {
#ifdef DEBUG
    printf("\n edge =%d (%ld %ld)\n",i,ta[i],ta[i+1]);
#endif   
    lp.p1.x = Mx(ta[i]);
    lp.p1.y = My(ta[i]);
    lp.p2.x = lp.p1.x;
    lp.p2.y = lp.p1.y;

    /* first check whether starting point is on test line - then ignore it */
    if ( !intersectl(lp,lt) ) {
      lp.p2.x = Mx(ta[j]);
      lp.p2.y =  My(ta[j]);
#ifdef DEBUG      
      printf("current i %d j %d count %d\n",i,j, count); 
#endif   
      
      j=i;
      if ( intersectl(lp,lt) )
	count +=1;
    }
#ifdef DEBUG
    else 
      printf("intersect with (%ld %ld) ",i,j);
#endif
  }
  i = 0; /* again tha first element */
#ifdef DEBUG 
  printf("\n edge =%d (%ld %ld)\n",i,ta[i],ta[i+1]);
#endif 
  lp.p1.x = Mx(ta[i]);
  lp.p1.y = My(ta[i]);
  lp.p2.x = lp.p1.x;
  lp.p2.y = lp.p1.y;
  
  /* first check whether starting point is on test line - then ignore it */
  if ( !intersectl(lp,lt) ) {
    lp.p2.x = Mx(ta[j]);
    lp.p2.y =  My(ta[j]);
#ifdef DEBUG   
    printf("current i %d j %d count %d\n",i,j, count);
#endif 
    j=i;
      if ( intersectl(lp,lt) )
	count +=1;
  }
#ifdef DEBUG
    else 
      printf("intersect with (%ld %ld) ",i,j);
#endif

#ifdef DEBUG
  printf("count %d\n", count); 
#endif  

  /* free memory */
  free(ta);

  if ( count%2 == 1 ) return( IN ); /* odd count */
  return( OUT ); /* even count */
}

//#define DEBUG

/* removes triangles outside the boundary */
/* here, we consider ALL boundaries */
/* we update material properties, parameters etc */
void 
cleanup_exterior(void)
{
  point pt;

  triag *t;
  unsigned long int i,j;
  

  /* for each boundary, we do this */
  /* first the outer boundary */
  for ( i = 1; i <= triangle_count; i++ ) {
    t = findTriangle(i);
    if ( t != 0 ) {
      /* check whether the centroid is inside */
      /* if not, delete the triangle */
      pt.x = (Mx(t->n1)+Mx(t->n2)+Mx(t->n3))*0.33333333333333333;
      pt.y = (My(t->n1)+My(t->n2)+My(t->n3))*0.33333333333333333;
      if ( inside(pt, boundarray[0].edge_count, boundarray[0].edges) == OUT ) {
	/* delete triangle */
#ifdef DEBUG
	printf("clean:deleting %ld  out 0\n",t->n);
#endif	
	deleteTriangle(*t);
      } else { /* update the properties of triangle */
#ifdef DEBUG
	printf("clean:keeping %ld  in 0\n",t->n);
#endif	
	t->boundary = 0;
	t->epsilon = boundarray[0].epsilon;
	
	/* division priority =1 */
	t->par = 1;
      }
    }
  }
  /* now we consider other boundaries if any */
  /* if boundary type = 0, remove triangles within boundary */
  for ( j = 1; j < nbound; j++ ) {
#ifdef DEBUG
      printf("clean:condsidering boundary %ld\n",j);
#endif
    for ( i = 1; i <= triangle_count; i++ ) {
#ifdef DEBUG
      printf("clean:condsidering tringle %ld\n",i);
#endif
      t = findTriangle(i);
      if ( t != 0 ) {
	/* check whether the centroid is inside */
	/* if not, delete the triangle */
	pt.x = (Mx(t->n1)+Mx(t->n2)+Mx(t->n3))*0.3333333;
	pt.y = (My(t->n1)+My(t->n2)+My(t->n3))*0.3333333;
	if ( inside(pt, boundarray[j].edge_count, boundarray[j].edges) == IN ) {
	  if ( boundarray[j].type == 0 ) {
	    /* delete */
#ifdef DEBUG
	    printf("clean:deleting %ld  out 0\n",t->n);
#endif	
	    deleteTriangle(*t);
	  } else { /* update the properties of triangle */
#ifdef DEBUG
	    printf("clean:IN boundary %ld ep = %Lf",j,  boundarray[j].epsilon);
#endif
	    t->boundary = j; /* t belongs to boundary j */
	    t->epsilon = boundarray[j].epsilon;
	  
	    /* instead of updating material, we could do */
	    /* with division priorities */
	  }
	}
      }
    }
  }
}
