/*  pdnmesh - a 2D finite element solver
    Copyright (c) 2001 Sarod Yatawatta <sarod@cs.pdn.ac.lk>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <stdio.h>
#include "types.h"



long double maximum_area; /* largest area of a triangle */
long double badness_limit; /* miimum badness for splitting */



/* a function to calculate the length of a line */
static long double 
len(unsigned long int p1, unsigned long int p2)
{
  long double temp;
  temp = (Mx(p1)-Mx(p2))*(Mx(p1)-Mx(p2))+(My(p1)-My(p2))*(My(p1)-My(p2));
  return(sqrt(temp));
}


/* a function to check whether a point p is on-line an edge e */
static int 
on_line(edge e, unsigned long int p)
{ 
  long double dx1,dx2,dy1,dy2;
  
  dx1 = Mx(e.p1)-Mx(p); 
#ifdef DEBUG 
  printf("dx1 %Lf ",dx1); 
#endif
  dy1 = My(e.p1)-My(p); 
#ifdef DEBUG 
  printf("dy1 %Lf ",dy1); 
#endif
  dx2 = Mx(p)-Mx(e.p2); 
#ifdef DEBUG 
  printf("dx2 %Lf ",dx2); 
#endif
  dy2 = My(p)-My(e.p2); 
#ifdef DEBUG 
  printf("dy2 %Lf\n",dy2); 
#endif
  
  /* first the coincident case */
  if ( ( fabsl(dx1) < TOL && fabsl(dy1) < TOL ) ||  ( fabsl(dx2) < TOL && fabsl(dy2) < TOL )) {
#ifdef DEBUG    
    printf("******************case1\n");
#endif
    return(1);
  }
  /* now to check parrallbility */
  /* we allow a toelarance, otherwise unpredictable  */
  if ( fabsl(dx1*dy2 - dx2*dy1) < TOL ) {
#ifdef DEBUG
    printf("__________________case2\n");
#endif    
    /* now check whether point is inbetween */
    /* this again gave problems, but when I changed */
    /* tha and to an or, and > to a < it worked */
    if ( (dx1*dx2 >= TOL && dy1*dy2 >=  TOL ) 
	 || ( (fabsl(dx1) < TOL) && (fabsl(dx2) < TOL ) && dy1*dy2 >= TOL )
	 || ( (fabsl(dy1) < TOL ) && (fabsl(dy2) < TOL ) && dx1*dx2 >= TOL ) )
	 return(1);
    return(0);
  }
  return(0);
}

 /* a function to check wether any line given by points p1,p2 is on */
/* any boundary edge - NOTE oth on same edge */
/* here we check only the INTERNAL boundaries */
static int 
on_boundary( unsigned long int p1, unsigned long int p2, unsigned long int *enumb )
{
  unsigned long int i;
#ifdef DEBUG
  printf("checking %ld %ld\n",p1,p2);
#endif 
  /* need only consider internal edges */
  for ( i = 0; i < bedges; i++ )
    if ( on_line(edgearray[i],p1) && on_line(edgearray[i],p2) ) {
#ifdef DEBUG
      printf("found %d!\n",i);
#endif     
      *enumb = i;
      return(1);
    }
  return(0);
}

/* a function to divide bad triangles */
int 
divide_bad_triangles(void)
{
  /* limiting factor to division */
  /* area > minimum and badness > worst */
  unsigned long int i,p1,p2,p3,p4,p,t1,t2,t3,t4;
  triag *t,*neigh;
  long double length1,length2,length3;
  point pt;
  triag new1,new2;
  int flag=0; /* flag to indicate of any division */
  // long double area_limit = 0.001; /* minimum area for splitting bad triangles */
 
  /* if this was exceeded, we split anyway */
  /* NOTE: we calculate the product of badness and area to decide */
  /* to split or not, THIS is stored in the above variable */
 
  long double badness,areat;
  unsigned long int old_triangle_count = triangle_count;
  long int enumb; /* edge number */ 

  for ( i = 1 ; i <= old_triangle_count; i++ ) {
    /* examine each triangle */
    t =(triag *)findTriangle(i);
#ifdef DEBUG
    printf("considering triangle %ld\n",i);
#endif   
    if ( t ) {
      length1 = len(t->n2,t->n3);
      length2 = len(t->n3,t->n1);
      length3 = len(t->n1,t->n2);
      /* we store the larges value in length1 */
      /* and the smalles value in length3     */
      /* we also keep track of the neightbour */
      /* sharing the largest edge */
      /* there are 3 possible cases */
      if ( (length1 >= length2) && (length1 >= length3 )) {
	/* this is what we want */
	if ( length3 > length2 )
	  length3 = length2;
	neigh=(triag *)findTriangle(t->t1);
	p1 = t->n1;
#ifdef DEBUG
      printf("case1\n");
#endif
      } else if ((length2 >= length1 ) && ( length2 >= length3 )) {
	if ( length1 > length3 )
	  length3 = length1;
	/* largest 2 */
	length1 = length2;
	neigh=(triag *)findTriangle(t->t2);
	p1 = t->n2;
#ifdef DEBUG
      printf("case2\n");
#endif
      } else if ( (length3 >= length1) && ( length3 >= length2 ) ) {
	if ( length2 > length1 )
	  length2 = length1;
	length1 = length3;
	length3 = length2;
	neigh=(triag *)findTriangle(t->t3);
	p1 = t->n3;
#ifdef DEBUG
      printf("case3\n");
#endif      
      } else { /* impossible case */
	continue; /* goto next traingle */
      } 
#ifdef DEBUG
      printf("longest %Lf shortest %Lf",length1,length3);
#endif      
#ifdef DEBUG
      printf("neighbour %ld\n",p1);
#endif
      /* only then should we think of division */
      /* we have a model like this             */
      /*               t.n1                    */
      /*               /\                      */
      /*              /  \                     */
      /*     length2 /    \ length3            */
      /*            /      \                   */
      /*           /        \                  */
      /*       t.n3----------t.n2              */
      /*             length1                   */
      badness = length1/length3;
      /* first calculate tha badness */
      /* now calculate the area */
      areat= area(*t);
      /* now depending upon the number of triangles, the neighbour, the area etc. */
      /* decide whether to split or not */
      /* here is the crucial desicion */
      /* I think for a triangle in tha boundary, we should only split the boundary edge */
      if (  neigh == 0 ) {
	if (  areat*t->par > maximum_area  
	      || badness > badness_limit ) {
	    /* now must split */
	  /* if neighbour is a boundary, split to two */
	  /*                   p1                      */
	  /*                  /\                       */
	  /*             t2  /| \ t1                   */
	  /*                / |  \                     */
	  /*               /  |p  \                    */
	  /*            p3 ---*---- p2                 */
	  /*               boundary                    */
	  /* boundary */
	  /* first determine the new point */
#ifdef DEBUG
	  printf("triangle %ld is bad, split to 2\n",t->n);
#endif
	  if ( p1 == t->n1 ) {
	    p2 = t->n2; t2 = t->t2;
	    p3 = t->n3; t1 = t->t3;
	  } else if ( p1 == t->n2 ) {
	    p2 = t->n3; t2 = t->t3;
	    p3 = t->n1; t1 = t->t1;
	  } else if (p1 == t->n3 ) {
	    p2 = t->n1; t2 = t->t1;
	    p3 = t->n2; t1 = t->t2;
	  }
	  /* now calculate the coordinates of point */
	  /* first the parameter */
	  /* a HORRIBLE expression indeed */
	
	  length1 = ((-My(p2)+My(p3))*(My(p1)-My(p2))-(Mx(p2)-Mx(p3))*(Mx(p1)-Mx(p2)))/((My(p1)-My(p3))*(My(p2)-My(p3))+(Mx(p1)-Mx(p3))*(Mx(p2)-Mx(p3)));
	  /* for our triangle, temp1 != -1 */
	  /* because we chose the largest edge */
	 
	  pt.x = (Mx(p2)+Mx(p3)*length1)/(1+length1);
	  pt.y = (My(p2)+My(p3)*length1)/(1+length1);
	  pt.pos = BOUNDARY;

	    /* interpolate to find the potential */
	  if ( fabsl(My(p3)-My(p2)) > TOL )
	    pt.pot = (pt.y-My(p2))*(Mpot(p3)-Mpot(p2))/(My(p3)-My(p2))+Mpot(p2);
	  else if (  fabsl(Mx(p3)-Mx(p2)) > TOL ) /* interpolate on x coordinates */
	    pt.pot = (pt.x-Mx(p2))*(Mpot(p3)-Mpot(p2))/(Mx(p3)-Mx(p2))+Mpot(p2);
	  else 
	    pt.pot = 0.5*(Mpot(p3)+Mpot(p2));
	  //	  printf("divide bad %Lf\n",pt.pot);
	  /* now insert point */
	  p =  insert(pt);
	  /* find the boundary as well */
	  if ( on_boundary(p,p, &enumb) > 0 ) {
#ifdef DEBUG
	    printf("intersec edge %ld\n",enumb);
#endif
	    if ( edgearray[enumb].type == DR )
	      Mval(p) = FX; /* dirichlet */
	    else 
	      Mval(p) = VAR; /* neumann */
	  }
	  new1.n = ++triangle_count;
	  new1.n1 = p1;
	  new1.n2 = p;
	  new1.n3 = p3;
	  new1.t1 = 0; /* BOUNDARY */
	  new1.t2 = t2;
	  new1.t3 = t->n;
	  new1.boundary = t->boundary;
	  /* evaluate rho for new1 */
#ifdef DEBUG
	    printf("bad1: boundary %ld\n",t->boundary);
#endif
	 
	  new1.epsilon = t->epsilon;
	  /* after division, reduce priority */
	  t->par /=2.0; /* divide by 2 */
	  /* copy priority to new atriangle */
	  new1.par =  t->par;
	  t->n1 = p1;
	  t->n2 = p2;
	  t->n3 = p;
	  t->t1 = 0; /* BOUNDARY */
	  t->t2 = new1.n;
	  t->t3 = t1;
	  /* evaluate rho for t */
#ifdef DEBUG
	    printf("bad2: boundary %ld\n",t->boundary);
#endif
	  
	  insertTriangle(new1);
	  flag =1; /* turn on flag */
#ifdef DEBUG
	  visitTree(root);
#endif
	}
	} else {
	/* neighbour exists */
	/* now we generalize according to the following */
	/*                           p1                  */
	/*                          /\                   */
	/*                         /  \                  */
	/*                     t2 /    \  t1             */
	/*                       /      \                */
	/*                      /        \               */
	/*                     /    *p    \              */
	/*                   p3------------p2            */
	/*                     \          /              */
	/*                      \  neigh /               */
	/*                       \      /                */
	/*                      t4\    / t3              */
	/*                         \  /                  */
	/*                          \/                   */
	/*                          p4                   */
	/*                           p1                  */
	/*                          /|                   */
	/*                         / |\                  */
	/*                     t2 /  | \  t1             */
	/*                       /   |  \                */
	/*                      /  t | new1              */
	/*                     /    *p    \              */
	/*                   p3------|-----p2            */
	/*                     \ neigh    /              */
	/*                      \    | new2              */
	/*                       \   |  /                */
	/*                      t4\  | / t3              */
	/*                         \ |/                  */
	/*                          \/                   */
	/*                          p4                   */
	/* only if neighbour's largest edge is common can we do anything */
	/* so calculate neighbour's largest edge */
	/* i.e. len(p2,p3) > len(p4,p3) and len(p2,p4)  */
#ifdef DEBUG
	  printf("triangle %ld is bad, neighbour %ld found\n",t->n,neigh->n);
#endif
	  
	  if ( p1 == t->n1 ) {
	    p2 = t->n2; t2 = t->t2;
	    p3 = t->n3; t1 = t->t3;
	  } else if ( p1 == t->n2 ) {
	    p2 = t->n3; t2 = t->t3;
	    p3 = t->n1; t1 = t->t1;
	  } else if (p1 == t->n3 ) {
	    p2 = t->n1; t2 = t->t1;
	    p3 = t->n2; t1 = t->t2;
	  }
	  /* now determine other detail as well */
	  if ( neigh->t1 == t->n ) {
	    p4 = neigh->n1;
	    if ( neigh->n2 == p2 ) { /* do we really need this? */
	      t4 = neigh->t2;
	    t3 = neigh->t3;
	    } else {
	      t4 = neigh->t3;
	    t3 = neigh->t2;
	    }
	} else if ( neigh->t2 == t->n ) {
	  p4 = neigh->n2;
	  if ( neigh->n1 == p2 ) {/* do we need to check? */
	    t4 = neigh->t1;
	    t3 = neigh->t3;
	  } else {
	    t4 = neigh->t3;
	    t3 = neigh->t1;
	  } 
	} else if ( neigh->t3 == t->n ) {
	  p4 = neigh->n3;
	  if ( neigh->n1 == p2 ) { /* do we need to check? */
	    t4 = neigh->t1;
	    t3 = neigh->t2;
	  } else {
	    t4 = neigh->t2;
	    t3 = neigh->t1;
	  } 
	} 
	
	
	  if ( /* (len(p2,p3) > len(p4,p3)) && (len(p2,p3) > len(p2,p4)) */
	      ccw2(p1,p4,p2) && ccw2(p1,p3,p4)
	       && (  areat*t->par > maximum_area 
		     || badness > badness_limit ))
	    
	    {
	    
	      
	      /* now determine the new point */
	      /* There is glitch here with internal edges */
	      /* if p2 and p3 are on a boundary, instead of */
	      /* centroid, we draw a normal to them */
	      if ((Mpos(p2) == BOUNDARY) &&  (Mpos(p3) == BOUNDARY)
		  && (on_boundary(p2,p3, &enumb) > 0) ) {
		
#ifdef DEBUG
		printf("edge number %ld\n",enumb);
#endif
		pt.pos = BOUNDARY;
		/* find any intersection */
		/* we take the normal to edge i */
		
		if ( edgearray[enumb].type == NU )
		  pt.val = VAR;
		else 
		  pt.val = FX;
		
		pt.x = (Mx(p2)+Mx(p3))*0.5;
		pt.y = (My(p2)+My(p3))*0.5;
		//		  printf("triagle %ld\n",t->n);	
		/* interpolate the potential , just divide by 2*/
		//pt.pot = (pt.y-(M.Narray[p2])->y)*((M.Narray[p3])->pot-(M.Narray[p2])->pot)/((M.Narray[p3])->y-(M.Narray[p2])->y)+(M.Narray[p2])->pot;
		pt.pot =0.5*(Mpot(p3)+Mpot(p2)); 
		/* this must be a material boundary */
		new1.boundary = t->boundary;
		new1.epsilon = t->epsilon;
		new2.boundary = neigh->boundary;
		new2.epsilon = neigh->epsilon;
	      } else  if ((Mpos(p1) == BOUNDARY) &&  (Mpos(p4) == BOUNDARY)  
			  && (on_boundary(p1,p4, &enumb) > 0)  )    {
#ifdef DEBUG
		printf("edge number %ld\n",enumb);
#endif	
		pt.pos = BOUNDARY;
		
		if ( edgearray[enumb].type == NU )
		  pt.val = VAR;
		else 
		  pt.val = FX;
		
		pt.x = (Mx(p1)+Mx(p4))*0.5;
		pt.y = (My(p1)+My(p4))*0.5;
		//printf("triagle %ld\n",t->n);		
		/* interpolate the potential just divide by 2*/
		//pt.pot = (pt.y-(M.Narray[p4])->y)*((M.Narray[p1])->pot-(M.Narray[p4])->pot)/((M.Narray[p1])->y-(M.Narray[p4])->y)+(M.Narray[p4])->pot;
		pt.pot =0.5*(Mpot(p4)+Mpot(p1)); 
		/* this can not be a material boundary */
		/* so we have one material only */
		new1.boundary = t->boundary;
		new1.epsilon = t->epsilon;
		new2.boundary = neigh->boundary;
		new2.epsilon = neigh->epsilon;
	      
	    }  else   {
	      
	      /* first take the centroid of the union */
	      pt.x = (Mx(p1)+Mx(p2)+Mx(p3)+Mx(p4))*0.25;
	      pt.y = (My(p1)+My(p2)+My(p3)+My(p4))*0.25;
	      pt.pos = IN;
	      pt.val = VAR;
	      /* if point is IN initialize potential to zero */
	      pt.pot = 0;
	      /* this can not be a material boundary */
	      /* so we have one material only */
	      new1.boundary = t->boundary;
	      new1.epsilon = t->epsilon;
	      new2.boundary = neigh->boundary;
	      new2.epsilon = neigh->epsilon;
	      
	    }
	  
	      // printf("divide bad 2 %Lf\n",pt.pot);
	      p =   insert(pt);
	     
	      /* now the new triangles */
	      new1.n = ++triangle_count;
	      new2.n = ++triangle_count;
	      new1.n1 = p1;
	      new1.n2 = p2;
	      new1.n3 = p;
	      new1.t1 = new2.n; 
	      new1.t2 = t->n;
	      new1.t3 = t1;
	      new2.n1 = p2;
	      new2.n2 = p4;
	      new2.n3 = p;
	      new2.t1 = neigh->n; 
	      new2.t2 = new1.n;
	      new2.t3 = t3;
	      /* evaluate rho for new1 */
#ifdef DEBUG
	    printf("bad3: boundary %ld ",new1.boundary);
#endif
	  
	   
#ifdef DEBUG
	    printf("bad3: rho %Lf\n", new1.rho);
#endif
	    
	      /* evaluate rho for new1 */
#ifdef DEBUG
	    printf("bad4: boundary %ld\n",new2.boundary);
#endif
	  
	      /* now modify old triangles */
	    t->n1 = p1;
	    t->n2 = p;
	    t->n3 = p3;
	    t->t1 = neigh->n;
	    t->t2 = t2;
	    t->t3 = new1.n;
	    /* evaluate rho for t */
#ifdef DEBUG
	    printf("bad5: boundary %ld\n",t->boundary);
#endif  
	   
	      neigh->n1 = p4;
	      neigh->n2 = p3;
	      neigh->n3 = p;
	      neigh->t1 = t->n;
	      neigh->t2 = new2.n;
	      neigh->t3 = t4;
	      /* evaluate rho for neigh */
#ifdef DEBUG
	    printf("bad6: boundary %ld\n",neigh->boundary);
#endif
	     
	      /* after division, reduce priority */
	      t->par =0.5*(t->par+neigh->par); /* divide by 2 */
	      /* copy priority to new atriangle */
	      new1.par =  new2.par = neigh->par = t->par;
	      /* now insert the new triangles */
	      insertTriangle(new1);
	      insertTriangle(new2);
	      flag = 1; /* turn on flag */
#ifdef DEBUG
	      visitTree(root);
#endif
	    }
	}
    }
  }
  return(flag);
}
  
  

/* a new function to split triangles with all 3 */
/* points at boundary */
void 
divide_corners(void)
{
  unsigned long int i, old_triangle_count,p,p1,p2,p3,p4,neigh1,neigh2,neigh3,neigh4;
  point pt;
  triag *t,*neigh;
  triag new1,new2;
  old_triangle_count = triangle_count;
  /* within the function, triangles increase, but need not */
  /* check new triangles */
  for ( i =1; i <= old_triangle_count; i++ ) {
    t =(triag *)findTriangle(i);
#ifdef DEBUG
    printf("considering triangle %ld\n",i);
#endif    /* Mval() == FX */
    if ( t && ( Mval(t->n1) == FX ) && ( Mval(t->n2) == FX ) && ( Mval(t->n3) == FX )) {
#ifdef DEBUG
      printf("splitting triangle %ld\n",i);
#endif  
      neigh = 0;
      /* we divide into three ONLY if no available neighbours are found */
      /* first need to find an available neighbour */
      if ( (t->t1 > 0 ) && (!on_boundary(t->n2,t->n3,&p ))) {
	neigh =(triag*)findTriangle(t->t1);
	p1 = t->n1;
	p2 = t->n2;
	p3 = t->n3;
	neigh1 = t->t2;
	neigh2 = t->t3;
      } else if ( (t->t2 > 0 ) && (!on_boundary(t->n1,t->n3, &p))) {
	neigh =(triag*)findTriangle(t->t2);
	p1 = t->n2;
	p2 = t->n3;
	p3 = t->n1;
	neigh1 = t->t3;
	neigh2 = t->t1;
      } else if ( (t->t3 > 0 ) && (!on_boundary(t->n1,t->n2, &p))) {
	neigh =(triag*)findTriangle(t->t3);
	p1 = t->n3;
	p2 = t->n1;
	p3 = t->n2;
	neigh1 = t->t1;
	neigh2 = t->t2;
      } else /* error */
	continue;
      if ( neigh ) { /* available neighbour */
	/*                                                            */
	/*                 p1                                         */
	/*                /\                                          */
	/*               /  \                                         */
	/*      neigh1  /    \ neigh2                                 */
	/*             /      \                                       */
	/*            /   t    \                                      */
	/*        p3 /__________\ p2                                  */
	/*           \          /                                     */
	/*            \   neigh/                                      */
	/*     neigh4  \      /  neigh3                               */
	/*              \    /                                        */
	/*               \  /                                         */
	/*                \/                                          */
	/*                p4                                          */
	
	/* generalise */
	if ( neigh->t1 == t->n ) {
	  p4 = neigh->n1;
	  neigh3 = neigh->t2;
	  neigh4 = neigh->t3;
	} else if ( neigh->t2 == t->n ) {
	  p4 = neigh->n2;
	  neigh3 = neigh->t3;
	  neigh4 = neigh->t1;
	} else if ( neigh->t3 == t->n ) {
	  p4 = neigh->n3;
	  neigh3 = neigh->t1;
	  neigh4 = neigh->t2;
	} 
	
	/* now find the midpoint of the points p2,p3 */
	pt.x = (Mx(p3)+Mx(p2))*0.5;
	pt.y = (My(p3)+My(p2))*0.5;
	pt.pos = IN; /* naturally */
	pt.val = VAR;
	pt.pot = 0; /* initialized to zero */
	/* insert point */
	p =  insert(pt); 
	/* new triangles */
	new1.n = ++triangle_count;
	new2.n = ++triangle_count;
	/*                                                            */
	/*                 p1                                         */
	/*                /\                                          */
	/*               /. \                                         */
	/*      neigh1  / .  \ neigh2                                 */
	/*             /  .   \                                       */
	/*            / t .new1\                                      */
	/*        p3 /____.pt___\ p2                                  */
	/*           \    `     /                                     */
	/*           neigh`new2/                                      */
	/*     neigh4  \  `   /  neigh3                               */
	/*              \ `  /                                        */
	/*               \` /                                         */
	/*                \/                                          */
	/*                p4                                          */
	/* this can not be a material boundary */
	/* so we have one material only */
	
	new1.epsilon = t->epsilon;
	
	new2.epsilon = neigh->epsilon;
	new1.n1 = p;
	new1.n2 = p1;
	new1.n3 = p2;
	new1.t1 = neigh2;
	new1.t2 = new2.n;
	new1.t3 = t->n;
	new2.n1 = p;
	new2.n2 = p2;
	new2.n3 = p4;
	new2.t1 = neigh3;
	new2.t2 = neigh->n ;
	new2.t3 = new1.n;
	new1.boundary = t->boundary;
	/* evaluate rho for new1 */
	
	new2.boundary = neigh->boundary;
	/* evaluate rho for new1 */
	
	t->n1 = p;
	t->n2 = p3;
	t->n3 = p1;
	t->t1 = neigh1;
	t->t2 = new1.n;
	t->t3 = neigh->n;
	neigh->n1 = p;
	neigh->n2 = p4;
	neigh->n3 = p3 ;
	neigh->t1 = neigh4;
	neigh->t2 = t->n;
	neigh->t3 = new2.n;
	/* evaluate rho for t */
	
	/* evaluate rho for neigh */
	
	/* after division, reduce priority */
	t->par =0.5*(t->par+neigh->par); /* divide by 2 */
	/* copy priority to new atriangle */
	new1.par =  new2.par = neigh->par = t->par;
	/* now insert triangles */
	insertTriangle(new1);
	insertTriangle(new2);
	
      } else { /* no choice, divide to three */

      /* first generalize */
      /*                      t.n1                                  */
      /*                      /\                                    */
      /*                     /  \                                   */
      /*             t.t3   /    \    t.t2 = neigh2                 */
      /*                   /      \                                 */
      /*                  /   *p new2                               */
      /*                 /          \                               */
      /*                /    new1    \                              */
      /*           t.n3 -------------- t.n2                         */
      /*                 neigh1                                     */  
      pt.x = (Mx(t->n1)+Mx(t->n2)+Mx(t->n3))*0.333333333;
      pt.y = (My(t->n1)+My(t->n2)+My(t->n3))*0.333333333;
      pt.pos = IN; /* naturally */
      pt.val =VAR;
      pt.pot = 0; /* initialized to zero */
      // printf("corners %Lf\n",pt.pot);
      p =  insert(pt); 
      /* new triangles */
      new1.n = ++triangle_count;
      new2.n = ++triangle_count;
      new1.n1 =p;
      new1.n2 =t->n2;
      new1.n3 =t->n3;
      new1.t1 =t->t1;
      new1.t2 =new2.n;
      new1.t3 =t->n;
      new2.n1 =p;
      new2.n2 =t->n3;
      new2.n3 =t->n1;
      new2.t1 =t->t2;
      new2.t2 =t->n;
      new2.t3 =new1.n;
      new1.boundary = t->boundary;
      /* evaluate rho for new1 */
     
      new2.boundary = t->boundary;
      /* evaluate rho for new1 */
     
      /* now adjust t */
      t->n3 =p;
      t->t1 =new1.n;
      t->t2 =new2.n;
      /* same material */
      new1.epsilon =  new2.epsilon = t->epsilon;
      /* evaluate rho for t */
     
      /* after division, reduce priority */
      t->par *=0.333333; /* divide by 3 */
      /* copy priority to new atriangle */
      new1.par =  new2.par =  t->par;
      /* now insert triangles */
      insertTriangle(new1);
      insertTriangle(new2);
      }
    }
  }
}



/* a function to determine whether point p4 is inside */
/* the circumcircle of points p1,p2,p3 */
static int 
inside_circle( unsigned long int p1, unsigned long int p2, unsigned long int p3, unsigned long int p4)
{
  long double temp1,temp2,temp3,temp4,a2,b2;
#ifdef DEBUG
  printf("checking (%ld,%ld,%ld) includes %ld\n",p1,p2,p3,p4);
#endif
  temp1 = (Mx(p1)*Mx(p1)-Mx(p2)*Mx(p2)+My(p1)*My(p1)-My(p2)*My(p2));
  temp2 = (Mx(p1)*Mx(p1)-Mx(p3)*Mx(p3)+My(p1)*My(p1)-My(p3)*My(p3));
  temp3 = (Mx(p1)-Mx(p2))*(My(p1)-My(p3))-(Mx(p1)-Mx(p3))*(My(p1)-My(p2));
   /* we know this is not zero */
  temp3 = 1/temp3;
  a2 = (temp1*(My(p1)-My(p3))-temp2*(My(p1)-My(p2)))*temp3;
  b2 = (temp2*(Mx(p1)-Mx(p2))-temp1*(Mx(p1)-Mx(p3)))*temp3;
  temp4 =  Mx(p1)*Mx(p1)-Mx(p4)*Mx(p4)+My(p1)*My(p1)-My(p4)*My(p4);
  
  if ( a2*(Mx(p1)-Mx(p4))+b2*(My(p1)-My(p4)) < temp4 ) {
#ifdef DEBUG
    printf("points are inside circle\n");
#endif
    return(1);
  }
  return(0);
}



/* function similar to splitting, but instead refine */
void 
delaunay_refine(void)
{
  unsigned long int i,p1,p2,p3,p4,t1,t2,t3,t4;
  triag *t,*neigh;
  long double length1,length2,length3;
  
  
  for ( i = 1 ; i <= triangle_count; i ++ ) {
    /* examine each triangle */
    t =(triag *)findTriangle(i);
#ifdef DEBUG
    printf("trying to refine triangle %ld\n",i);
#endif   
    if ( t ) {
      length1 = len(t->n2,t->n3);
      length2 = len(t->n3,t->n1);
      length3 = len(t->n1,t->n2);
      /* we store the larges value in length1 */
      /* and the smalles value in length3     */
      /* we also keep track of the neightbour */
      /* sharing the largest edge */
      /* there are 3 possible cases */
      if ( (length1 > length2) && (length1 > length3 )) {
	/* this is what we want */
	if ( length3 > length2 )
	  length3 = length2;
	neigh=(triag *)findTriangle(t->t1);
	p1 = t->n1;
#ifdef DEBUG
      printf("case1 ");
#endif
      } else if ((length2 > length1 ) && ( length2 > length3 )) {
	if ( length1 > length3 )
	  length3 = length1;
	/* largest 2 */
	length1 = length2;
	neigh=(triag *)findTriangle(t->t2);
	p1 = t->n2;
#ifdef DEBUG
      printf("case2 ");
#endif
      } else if ( (length3 > length1)&& ( length3 > length2 ) ) {
	if ( length2 > length1 )
	  length2 = length1;
	length1 = length3;
	length3 = length2;
	neigh=(triag *)findTriangle(t->t3);
	p1 = t->n3;
#ifdef DEBUG
      printf("case3 ");
#endif      
      } else {
	/* the case when lengths are equal */
	continue;
#ifdef DEBUG
      printf("lengths are equal\n");
#endif    
      }
      /* now neighbour is neigh */
#ifdef DEBUG
      printf("longest %Lf shortest %Lf\n",length1,length3);
#endif      

      if ( neigh  ) {
#ifdef DEBUG
      printf("neighbour %ld ",neigh->n);
#endif
	/* neighbour exists */
	/* now we generalize according to the following */
	/*                           p1                  */
	/*                          /\                   */
	/*                         /  \                  */
	/*                     t2 /    \  t1             */
	/*                       /      \                */
	/*                      /        \               */
	/*                     /          \              */
	/*                   p3------------p2            */
	/*                     \          /              */
	/*                      \  neigh /               */
	/*                       \      /                */
	/*                      t4\    / t3              */
	/*                         \  /                  */
	/*                          \/                   */
	/*                          p4                   */
	/*                           p1                  */
	/*                          /|                   */
	/*                         / |\                  */
	/*                     t2 /  | \  t1             */
	/*                       /   |  \                */
	/*                      /    |   \               */
	/*                     /     |    \              */
	/*                   p3    t |     p2            */
	/*                     \      neigh              */
	/*                      \    |   /               */
	/*                       \   |  /                */
	/*                      t4\  | / t3              */
	/*                         \ |/                  */
	/*                          \/                   */
	/*                          p4                   */
	/* only if neighbour's largest edge is common can we do anything */
	/* so calculate neighbour's largest edge */
	/* i.e. len(p2,p3) > len(p4,p3) and len(p2,p4)  */
#ifdef DEBUG
	  printf(" triangle %ld is bad, neighbour %ld found\n",t->n,neigh->n);
#endif
	  if ( p1 == t->n1 ) {
	    p2 = t->n2; t2 = t->t2;
	    p3 = t->n3; t1 = t->t3;
	  } else if ( p1 == t->n2 ) {
	    p2 = t->n3; t2 = t->t3;
	    p3 = t->n1; t1 = t->t1;
	  } else if (p1 == t->n3 ) {
	    p2 = t->n1; t2 = t->t1;
	    p3 = t->n2; t1 = t->t2;
	  }
	  /* now determine other detail as well */
	  if ( neigh->t1 == t->n ) {
	    p4 = neigh->n1;
	    if ( neigh->n2 == p2 ) { /* do we really need this? */
	      t4 = neigh->t2;
	    t3 = neigh->t3;
	    } else {
	      t4 = neigh->t3;
	    t3 = neigh->t2;
	    }
	} else if ( neigh->t2 == t->n ) {
	  p4 = neigh->n2;
	  if ( neigh->n1 == p2 ) {/* do we need to check? */
	    t4 = neigh->t1;
	    t3 = neigh->t3;
	  } else {
	    t4 = neigh->t3;
	    t3 = neigh->t1;
	  } 
	} else if ( neigh->t3 == t->n ) {
	  p4 = neigh->n3;
	  if ( neigh->n1 == p2 ) { /* do we need to check? */
	    t4 = neigh->t1;
	    t3 = neigh->t2;
	  } else {
	    t4 = neigh->t2;
	    t3 = neigh->t1;
	  } 
	}
	  
	  /* ONLY if the new badnesses of both triangles together are less than the earlier one */
	  /* we do the change */
	  /* OR in other words, only if p4 is inside the circumcircle of p1,p2,p3 */
	  /* and we DON NOT refine triangles at corners that are already divided */

	  if ( inside_circle(p1,p2,p3,p4) 
	       //&& Mpos(p1) == IN && Mpos(p4) == IN
	       //&& Mpos(p2) != BOUNDARY && Mpos(p3) != BOUNDARY) {
	       /* here we do not need variable length1 */
	       && !on_boundary(p2,p3,&length1) 
	       && !((Mval(p1) == FX) && (Mval(p2) == FX) && (Mval(p4) == FX) ) 
	       && !((Mval(p1) == FX) && (Mval(p3) == FX) && (Mval(p4) == FX) ) ) {
	       /* not on the same boundary(p2,p3) */
	    /* now modify old triangles */
	    t->n1 = p1;
	    t->n2 = p4;
	    t->n3 = p3;
	    t->t1 = t4;
	    t->t2 = t2;
	    t->t3 = neigh->n;
	    neigh->n1 = p4;
	    neigh->n2 = p1;
	    neigh->n3 = p2;
	    neigh->t1 = t1;
	    neigh->t2 = t3;
	    neigh->t3 = t->n;
	    /* now DO NOT forget to update the neighbouring triangles */
	    /* to this pair */
	    /* use the same pointer as neigh to address them , but */
	    /* it should not be misunderstood */
	    if ( t1 ) {
	      neigh = (triag *)findTriangle(t1);
	      if ( neigh ) {
		if ( neigh->t1 == t->n )
		  neigh->t1 = t->t3; /* which is equal to neighbour */
		else if ( neigh->t2 == t->n )
		  neigh->t2 = t->t3;
		else if ( neigh->t3 == t->n )
		  neigh->t3 = t->t3;
	      }
	    }
	    /* do this for t4 as well */
	    if ( t4 ) {
	      neigh = (triag *)findTriangle(t4);
	      if ( neigh ) {
		if ( neigh->t1 == t->t3 )
		  neigh->t1 = t->n; /* which is equal to neighbour */
		else if ( neigh->t2 == t->t3 )
		  neigh->t2 = t->n;
		else if ( neigh->t3 == t->t3 )
		  neigh->t3 = t->n;
	      }
	    }
	    
#ifdef DEBUG
	    visitTree(root);
#endif
	  }
      }
    }
  }
}
